import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class ICharTokenizer(ABC):

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def peek(self) -> str:
        pass

    @abstractmethod
    def read(self) -> str:
        pass

    @abstractmethod
    def unread(self, char: str) -> str:
        pass

    @abstractmethod
    def get_pos(self) -> int:
        pass

    @abstractmethod
    def get_segment(self, offset: int, length: int) -> str:
        pass



import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class ICsvConsumer(ABC):

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def consume_field(self, s: str) -> None:
        pass

    @abstractmethod
    def signal_end_of_record(self) -> None:
        pass

    @abstractmethod
    def signal_end_of_line(self) -> None:
        pass



from __future__ import annotations

import logging

from csvlineparser.csv_console_consumer import CsvConsoleConsumer
from csvlineparser.csv_context_provider import CsvContextProvider

logger = logging.getLogger(__name__)


class CsvCleanerConsumer(CsvConsoleConsumer):

    def __init__(self, has_header: bool = True, csv_context_provider: CsvContextProvider | None = None) -> None:
        super().__init__()
        self.csv_context_provider = csv_context_provider
        if csv_context_provider:
            self.delimiter = csv_context_provider.get_delimiter()
        else:
            self.delimiter = ','
        self.has_header = has_header
        self.__line = ''
        self.__line_num = 0
        self.__col_count = 0
        self.__all_lines = []

    def consume_field(self, input_value: str) -> None:
        super().consume_field(input_value)

        if self.csv_context_provider:
            if self.__line_num == 0:
                if self.has_header:
                    input_value = self.csv_context_provider.handle_header_element(input_value, self.__line_num,
                                                                                  self.__col_count)
                else:
                    input_value = self.csv_context_provider.handle_data_element(input_value, self.__line_num, self.__col_count)
            else:
                input_value = self.csv_context_provider.handle_data_element(input_value, self.__line_num, self.__col_count)

        self.__line += input_value + self.delimiter
        self.__col_count += 1

    def signal_end_of_record(self) -> None:
        super().signal_end_of_record()
        self.__line = self.__line[:-1]
        self.__all_lines.append(self.__line)
        self.__line = ''
        self.__line_num += 1
        self.__col_count = 0

    def signal_end_of_line(self) -> None:
        super().signal_end_of_line()

    def get_header(self) -> str:
        return self.__all_lines[0]

    def get_data(self) -> list:
        return self.__all_lines[1:]

    def get_all(self) -> list:
        return self.__all_lines

    def get_line(self, line_num: int) -> str:
        return self.__all_lines[line_num]



from __future__ import annotations

import logging

from csvlineparser.exceptions import IllegalArgumentException

logger = logging.getLogger(__name__)


class CsvDataTypeMapping:
    """ Put this into a map with key being column name.
    If you clean up the key, use the original one. """

    def __init__(self, col_name: str, data_type: str, required: bool, ignored: bool) -> None:
        super().__init__()
        if data_type not in ['double', 'int', 'boolean', 'string', 'datetime']:
            raise IllegalArgumentException(f'Currently not supported data type for CSV cleaning: {data_type}')
        self.__col_name = col_name
        self.__data_type = data_type
        self.__required = required
        self.__ignored = ignored

    def get_col_name(self) -> str:
        """ Returns the name, for convenience, mostly debugging and choking up """
        return self.__col_name

    def get_type(self) -> str:
        """ currently only: double, int, boolean, string, datetime """
        return self.__data_type

    def is_required(self) -> bool:
        """ Will either a) throw an exception if isIgnored() is true, else b) log a warning """
        return self.__required

    def is_ignored(self) -> bool:
        return self.__ignored

    def __repr__(self) -> str:
        return 'n/a'



import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class CsvContextProvider(ABC):

    def __init__(self) -> None:
        super().__init__()

    def get_delimiter(self) -> str:
        """ Returns the char used to delimit columns. Defaults to comma (,) """
        return ','

    @abstractmethod
    def handle_header_element(self, input_value: str, row: int, column: int) -> str:
        """ Apply transformations to the input value, which is a header cell.
        The row and column parameters are there if you need it. Return value
        will be inserted into resulting CSV object """
        pass

    @abstractmethod
    def handle_data_element(self, input_value: str, row: int, column: int) -> str:
        """ Apply transformations to the input value, which is a data cell.
        The row and column parameters are there if you need it. Return value
        will be inserted into resulting CSV object """
        pass



from __future__ import annotations

import logging

from csvlineparser.csv_consumer_interface import ICsvConsumer
from csvlineparser.exceptions import IllegalStateException

logger = logging.getLogger(__name__)


class MultiConsumer(ICsvConsumer):

    def __init__(self, fail_fast: bool = False) -> None:
        """ ICsvConsumer can be appended, and it will delegate the call to many.
        If failFast, will raise an exception data columns differ from header columns
        (or generally the first column) """
        super().__init__()
        self.__header_read: bool = False
        self.__header_elem_count = 0
        self.__line_elem_count = 0
        self.__consumers: list[ICsvConsumer] = []
        self.__fail_fast: bool = fail_fast
        self.__line_num = 1

    def addConsumers(self, consumer: ICsvConsumer | list[ICsvConsumer]) -> None:
        if isinstance(consumer, list):
            self.__consumers.extend(consumer)
        else:
            self.__consumers.append(consumer)

    def consume_field(self, s: str) -> None:
        if not self.__header_read:
            self.__header_elem_count += 1
        else:
            self.__line_elem_count += 1
        for consumer in self.__consumers:
            consumer.consume_field(s)

    def signal_end_of_record(self) -> None:
        if self.__header_read:
            if self.__header_elem_count != self.__line_elem_count:
                if self.__fail_fast:
                    raise IllegalStateException(f'Header element count [{self.__header_elem_count}] does not '
                                    f'match line element count [{self.__line_elem_count}]. '
                                    f'Line: {self.__line_num}')
        self.__header_read = True
        self.__line_elem_count = 0
        self.__line_num += 1
        for consumer in self.__consumers:
            consumer.signal_end_of_record()

    def signal_end_of_line(self) -> None:
        for consumer in self.__consumers:
            consumer.signal_end_of_line()



from __future__ import annotations

import logging

from csvlineparser.char_tokenizer_interface import ICharTokenizer
from csvlineparser.csv_consumer_interface import ICsvConsumer
from csvlineparser.exceptions import CsvParserTooMuchDataException, CsvParserNoTermQuoteException
from csvlineparser.multi_consumer import MultiConsumer

logger = logging.getLogger(__name__)


class CsvParser:

    def __init__(self, debug_offset: int = 20) -> None:
        super().__init__()
        self.__first_row: bool = True
        self.__current_row: str = ''
        self.__debug_offset: int = debug_offset
        self.__line_number: int = 0

    def parse(self, reader: ICharTokenizer,
              consumers: ICsvConsumer | list[ICsvConsumer],
              fail_fast: bool = True) -> None:
        multi_consumer = MultiConsumer(fail_fast=fail_fast)
        multi_consumer.addConsumers(consumers)
        self.__parse_csv_file(reader, multi_consumer)

    def __parse_csv_file(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        while reader.peek() != 'EOF':
            self.__parse_csv_record(reader, consumer)
        consumer.signal_end_of_line()

    def __parse_csv_record(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        self.__line_number += 1
        self.__parse_csv_string_list(reader, consumer)
        ch: str = reader.read()
        if ch == 'EOF':
            reader.unread(ch)
            ch = '\n'
        if ch != '\n':
            raise CsvParserTooMuchDataException(f'End of record was expected but more data exists: >{ch}<',
                                                reader.get_pos(), reader.get_segment(reader.get_pos()
                                                                                     - self.__debug_offset,
                                                                                     self.__debug_offset * 2),
                                                self.__debug_offset, self.__line_number)
        consumer.signal_end_of_record()

    def __parse_csv_string_list(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        ch: str
        while True:
            self.__parse_raw_string(reader, consumer)
            ch = reader.read()
            if ch != ',':
                break
        reader.unread(ch)

    def __is_field_terminator(self, c: str) -> bool:
        return c == ',' or c == '\n' or c == 'EOF'

    def __is_space(self, c: str) -> bool:
        return c == ' ' or c == '\t'

    def __parse_optional_space(self, reader: ICharTokenizer) -> None:
        ch: str
        while True:
            ch = reader.read()
            if not self.__is_space(ch):
                break
        reader.unread(ch)

    def __parse_raw_string(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        self.__parse_optional_space(reader)
        self.__parse_raw_field(reader, consumer)
        if not self.__is_field_terminator(reader.peek()):
            self.__parse_optional_space(reader)

    def __parse_raw_field(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        field_value: str = ''

        ch: str = reader.peek()
        if not self.__is_field_terminator(ch):
            if ch == '"':
                field_value = self.__parse_quoted_field(reader)
            else:
                field_value = self.__parse_simple_field(reader)
        consumer.consume_field(field_value.strip())

    def __parse_quoted_field(self, reader: ICharTokenizer) -> str:
        reader.read()  # read and discard initial quote

        field: str = self.__parse_escaped_field(reader)

        ch: str = reader.read()
        if ch != '"':
            reader.unread(ch)
            raise CsvParserNoTermQuoteException('Quoted field has no terminating double quote')
        return field

    def __parse_escaped_field(self, reader: ICharTokenizer) -> str:
        sb: list[str] = list()

        self.__parse_sub_field(reader, sb)
        ch: str = reader.read()
        while self.__process_double_quote(reader, ch):
            sb.append('"')
            self.__parse_sub_field(reader, sb)
            ch = reader.read()
        reader.unread(ch)

        return ''.join(sb)

    def __parse_sub_field(self, reader: ICharTokenizer, sb: list[str]) -> None:
        ch: str = reader.read()
        while ch != '"' and ch != 'EOF':
            sb.append(ch)
            ch = reader.read()
        reader.unread(ch)

    def __is_bad_simple_field_char(self, c: str) -> bool:
        return self.__is_field_terminator(c) or c == '"'

    def __parse_simple_field(self, reader: ICharTokenizer) -> str:
        ch: str = reader.read()
        if self.__is_bad_simple_field_char(ch) or self.__is_space(ch):
            reader.unread(ch)
            return ''

        sb: list[str] = list()
        sb.append(ch)
        ch = reader.read()
        while not self.__is_bad_simple_field_char(ch):
            sb.append(ch)
            ch = reader.read()
        reader.unread(ch)
        return ''.join(sb)

    def __process_double_quote(self, reader: ICharTokenizer, ch: str) -> bool:
        if ch == '"' and reader.peek() == '"':
            reader.read()  # discard second quote of double
            return True
        return False



import logging

logger = logging.getLogger(__name__)


class CharTokenizerException(Exception):

    def __init__(self, message: str):
        super().__init__(self, message)


class CsvParserTooMuchDataException(Exception):

    def __init__(self, message: str, pos: int, data: str, offset: int, line_number: int):
        super().__init__(self, message)
        data = data.replace('\r', '\\r').replace('\n', '\\n')
        print(f'line: {line_number}, pos: {pos}')
        print(f'{data}')
        print(f'{"-" * offset}^')


class CsvParserNoTermQuoteException(Exception):

    def __init__(self, message: str):
        super().__init__(self, message)


class IllegalStateException(Exception):

    def __init__(self, message: str):
        super().__init__(self, message)


class IllegalArgumentException(Exception):

    def __init__(self, message: str):
        super().__init__(self, message)



import logging
from abc import ABC

from csvlineparser.char_tokenizer_interface import ICharTokenizer
from csvlineparser.exceptions import CharTokenizerException

logger = logging.getLogger(__name__)


class StringCharTokenizer(ICharTokenizer, ABC):

    def __init__(self, s: str) -> None:
        super().__init__()
        self.__s: str = s
        self.__index: int = 0
        self.__have_unread_char: bool = False
        self.__unread_char: str = ''

    def __skipCrInCrLf(self) -> None:
        if self.__s[self.__index] == '\r' and self.__index + 1 < len(self.__s) and self.__s[self.__index + 1] == '\n':
            self.__index += 1

    def __mapCrToLf(self, c: str) -> str:
        if c == '\r':
            return '\n'
        return c

    def peek(self) -> str:
        if self.__have_unread_char:
            return self.__unread_char
        if self.__index < len(self.__s):
            return self.__mapCrToLf(self.__s[self.__index])
        return 'EOF'

    def read(self) -> str:
        if self.__have_unread_char:
            self.__have_unread_char = False
            return self.__unread_char
        if self.__index < len(self.__s):
            self.__skipCrInCrLf()
            ret_val = self.__mapCrToLf(self.__s[self.__index])
            self.__index += 1
            return ret_val
        return 'EOF'

    def unread(self, c: str) -> None:
        if self.__have_unread_char:
            raise CharTokenizerException('unread() cannot accept more than one pushed back character')
        self.__have_unread_char = True
        self.__unread_char = c

    def get_pos(self) -> int:
        return self.__index

    def get_segment(self, offset: int, length: int) -> str:
        return self.__s[offset:offset + length]



import logging

from csvlineparser.csv_consumer_interface import ICsvConsumer

logger = logging.getLogger(__name__)


class CsvConsoleConsumer(ICsvConsumer):

    def __init__(self) -> None:
        super().__init__()
        self.__colIndex = 0
        self.__rowIndex = 0
        self.__num_cols = 0
        self.__num_rows = 0
        self.__matrix = {}

    def consume_field(self, s: str) -> None:
        self.__matrix[self.__colIndex, self.__rowIndex] = s
        self.__colIndex += 1
        if self.__num_cols < self.__colIndex:
            self.__num_cols += 1

    def signal_end_of_record(self) -> None:
        self.__rowIndex += 1
        self.__num_rows += 1
        self.__colIndex = 0

    def signal_end_of_line(self) -> None:
        print(f'End of file reached. Got {self.__num_cols} columns and {self.__num_rows} rows')

    def get_matrix(self) -> dict:
        return self.__matrix

    def get_column(self, col_index: int) -> list:
        col = []
        for i in range(0, self.__num_rows):
            col.append(self.__matrix[col_index, i])
        return col

    def get_row(self, row_index: int) -> list:
        row = []
        for i in range(0, self.__num_cols):
            row.append(self.__matrix[i, row_index])
        return row

    def get_num_rows(self) -> int:
        return self.__num_rows

    def get_num_cols(self) -> int:
        return self.__num_cols



from __future__ import annotations

import datetime
import logging

from csvlineparser.csv_context_provider import CsvContextProvider
from csvlineparser.csv_data_type_mapping import CsvDataTypeMapping

logger = logging.getLogger(__name__)


class AwareCsvContextProvider(CsvContextProvider):

    def __init__(self, data_type_mappings: dict[int, CsvDataTypeMapping] | None) -> None:
        super().__init__()
        self.__lineNum: int = 0
        self.__colCount: int = 0
        self.__refData: dict[int, CsvDataTypeMapping] | None = data_type_mappings

    def get_delimiter(self) -> str:
        """ Returns the char used to delimit columns. Defaults to comma (,) """
        return ','

    def handle_header_element(self, input_value: str, row: int, column: int) -> str:
        """ Apply transformations to the input value, which is a header cell.
        The row and column parameters are there if you need it. Return value
        will be inserted into resulting CSV object """
        # header for glue cannot have dots (.)
        if '.' in input_value:
            input_value = input_value.replace('.', '_')
        return input_value

    def handle_data_element(self, input_value: str, row: int, column: int) -> str:
        """ Apply transformations to the input value, which is a data cell.
        The row and column parameters are there if you need it. Return value
        will be inserted into resulting CSV object """
        if self.__lineNum == 0 and '.' in input_value:
                input_value = input_value.replace('.', '_')
        if input_value != '':
            if input_value.lower() == 'true' or input_value.lower() == 'false':
                input_value = input_value.lower()
            if '"' in input_value:
                input_value = input_value.replace('"', '""')
            if self.__toDate(input_value) is not None:
                input_value = self.__toDate(input_value)
            elif self.get_delimiter() in input_value:
                input_value = f'"{input_value}"'
        else:
            input_value = ''

        if self.__refData:
            expected_data_type = self.__refData[self.__colCount].get_type()
            if input_value == '' and self.__refData[self.__colCount].is_required():
                if expected_data_type == 'double':
                    logger.warning(f'Empty double found in column: {self.__refData[self.__colCount].get_col_name()}. '
                                   f'Replacing with 0.0')
                    input_value = '0.0'
                elif expected_data_type == 'int':
                    logger.warning(f'Empty int found in column: {self.__refData[self.__colCount].get_col_name()}. '
                                   f'Replacing with 0')
                    input_value = '0'
                elif expected_data_type == 'boolean':
                    logger.warning(f'Empty boolean found in column: {self.__refData[self.__colCount].get_col_name()}. '
                                   f'Replacing with False')
                    input_value = 'false'
                elif expected_data_type == 'datetime':
                    raise ValueError(
                        f'Empty date datetime found for column: {self.__refData[self.__colCount].get_col_name()}')
                    # logger.warning(f'Empty date datetime found for column: {self.refData[self.rowCount].name}')
                elif expected_data_type == 'string':
                    logger.info(f'Skipping: Empty string: {self.__refData[self.__colCount].get_col_name()}, '
                                f'expected type: {expected_data_type}')
                else:
                    raise ValueError(f'Empty value found for column: {self.__refData[self.__colCount].get_col_name()} '
                                     f'and type: {expected_data_type}')
        return input_value

    def __toDate(self, s: str) -> str | None:
        """ Check if string should be a date of format: YYYY-MM-DD or YYYY-MM-DD hh:mm:ss """
        try:
            return datetime.datetime.strptime(s, "%Y-%m-%d").strftime("%Y-%m-%d")
        except ValueError:
            pass
        try:
            return datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S")
        except ValueError:
            pass
        try:
            return datetime.datetime.strptime(s, "%d/%m/%Y").strftime("%Y-%m-%d")
        except ValueError:
            pass
        try:
            return datetime.datetime.strptime(s, "%d/%m/%Y %H:%M").strftime("%Y-%m-%d %H:%M:%S")
        except ValueError:
            pass
        try:
            return datetime.datetime.strptime(s, "%d/%m/%Y %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S")
        except ValueError:
            pass
        return None



import logging

from csvlineparser.csv_consumer_interface import ICsvConsumer
from csvlineparser.exceptions import IllegalStateException

logger = logging.getLogger(__name__)


class CsvConsumerTester(ICsvConsumer):

    def __init__(self, expected_fields: list[str]) -> None:
        super().__init__()
        self.expected_fields: list[str] = expected_fields
        self.field_index: int = 0

    def consume_field(self, s: str) -> None:
        if self.expected_fields[self.field_index] != s:
            raise IllegalStateException(f'field [{self.expected_fields[self.field_index]}] expected, but [{s}] returned')
        self.field_index += 1

    def signal_end_of_record(self) -> None:
        if self.expected_fields[self.field_index] != 'EOR':
            self.field_index += 1
            raise IllegalStateException('End of record signalled but not expected')

    def signal_end_of_line(self) -> None:
        if self.expected_fields[self.field_index] != '':
            self.field_index += 1
            raise IllegalStateException('End of file signalled but not expected')
